import collections
import copy
import warnings

from typing import Optional, List, Dict

from ..model import Event, InternalEvent, MetaEvent


__all__ = ['TimeContextProvider', 'EventContextProvider', 'FrozenContext']


warnings.warn('sismic.code.context is deprecated since Sismic 1.4.1', DeprecationWarning)


class TimeContextProvider:  # pragma: no cover
    """
    A context provider for time related predicates.

    This context exposes time, after, idle, and active.
    Look at their respective documentation for more information.

    This provider needs to be attached to an interpreter.
    """

    def __init__(self) -> None:
        self._entry_time = dict()  # type: Dict[str, float]
        self._idle_time = dict()  # type: Dict[str, float]
        self._time = 0  # type: float
        self._configuration = []  # type: List[str]

    @property
    def time(self) -> float:
        """
        Current time of the interpreter.
        """
        return self._time

    def after(self, name: str, seconds: float) -> bool:
        """
        Return True if and only if given state was entered for more than given
        time, expressed in seconds.

        :param name: name of the state.
        :param seconds: elapsed time to use for comparison.
        :return: True iff. given state was entered for more than given time.
        """
        return self.time - seconds >= self._entry_time[name]

    def idle(self, name: str, seconds: float) -> bool:
        """
        Return True if and only if given state did not fire a transition for
        more than given time, expressed in seconds.

        :param name: name of the state.
        :param seconds: elapsed time to use for comparison.
        :return: True iff. given state did not fire a transition for more than given time.
        """
        return self.time - seconds >= self._idle_time[name]

    def active(self, name: str) -> bool:
        """
        Return True if and only if given state is active.

        :param name: name of the state.
        :return: True iff. given state is active.
        """
        return name in self._configuration

    def __call__(self, event: MetaEvent):
        if event.name == 'step started':
            self._time = event.time
        elif event.name == 'state entered':
            self._configuration.append(event.state)
            self._entry_time[event.state] = self._time
            self._idle_time[event.state] = self._time
        elif event.name == 'state exited':
            self._configuration.remove(event.state)
        elif event.name == 'transition processed':
            self._idle_time[event.source] = self._time


class EventContextProvider:  # pragma: no cover
    """
    A context provider for event related predicates.

    This context exposes send, notify, sent and received.
    Look at their respective documentation for more information.

    The list of events that were sent during the last executed step is
    available through the ``pending`` attribute. This list should be returned
    by the evaluator on code execution for the events to be effectively sent.

    This provider needs to be attached to an interpreter.
    """

    def __init__(self) -> None:
        self.pending = []  # type: List[Event]
        self._sent = []  # type: List[Event]
        self._consumed = None  # type: Optional[Event]

    def send(self, name: str, **kwargs) -> None:
        """
        Create an internal event and store it for further sending.

        :param name: name of the event.
        :param **kwargs: additional event parameters.
        """
        self.pending.append(InternalEvent(name, **kwargs))

    def notify(self, name: str, **kwargs) -> None:
        """
        Create a meta event and store it for further sending.

        :param name: name of the event.
        :param **kwargs: additional event parameters.
        """
        self.pending.append(MetaEvent(name, **kwargs))

    def sent(self, name: str) -> bool:
        """
        Return True if and only if given internal event was sent during
        current step.

        :param name: name of the event.
        :return: True iff. event was sent.
        """
        return any((name == e.name for e in self._sent))

    def received(self, name: str) -> bool:
        """
        Return True if and only if given event is currently processed.
        This function should only be used during contract evaluation, and not
        during guard evaluation, as it relies on the *consumed* event.

        :param name: name of the event.
        :return: True iff. event is processed.
        """
        return getattr(self._consumed, 'name', None) == name

    def __call__(self, event: MetaEvent) -> None:
        if event.name == 'event consumed':
            self._consumed = event.event
        elif event.name == 'event sent':
            self._sent.append(event.event)
        elif event.name == 'step started':
            self._consumed = None
            self._sent = []
            self.pending = []


class FrozenContext(collections.abc.Mapping):  # pragma: no cover
    """
    A shallow copy of a context. The keys of the underlying context are
    exposed as attributes.
    """
    __slots__ = ['__frozencontext']

    def __init__(self, context: Dict) -> None:
        self.__frozencontext = {k: copy.copy(v) for k, v in context.items()}

    def __getattr__(self, item):
        try:
            return self.__frozencontext[item]
        except KeyError:
            raise AttributeError('{} has no attribute {}'.format(self, item))

    def __getstate__(self):
        return self.__frozencontext

    def __setstate__(self, state):
        self.__frozencontext = state

    def __getitem__(self, key):
        return self.__frozencontext[key]

    def __len__(self):
        return len(self.__frozencontext)

    def __iter__(self):
        return iter(self.__frozencontext)



from behave import given, when, then  # type: ignore
from .. import testing


@given('I do nothing')
@when('I do nothing')
def do_nothing(context):
    pass


@given('I reproduce "{scenario}"')
def reproduce_scenario(context, scenario, *, keyword='Given'):
    current_feature = context.feature
    for included_scenario in current_feature.scenarios:
        if included_scenario.name == scenario:
            for step in included_scenario.steps:
                if step.step_type in ['given', 'when']:
                    context.execute_steps('{} {}'.format(keyword, step.name))
            return
    assert False, 'Unknown scenario {}.'.format(scenario)


@when('I reproduce "{scenario}"')
def _reproduce_scenario(context, scenario):
    return reproduce_scenario(context, scenario, keyword='When')


@given('I repeat "{step}" {repeat:d} times')
def repeat_step(context, step, repeat, *, keyword='Given'):
    for _ in range(repeat):
        context.execute_steps('{} {}'.format(keyword, step))


@when('I repeat "{step}" {repeat:d} times')
def _repeat_step(context, step, repeat):
    return repeat_step(context, step, repeat, keyword='When')


@given('I send event {name}')
@given('I send event {name} with {parameter}={value}')
@when('I send event {name}')
@when('I send event {name} with {parameter}={value}')
def send_event(context, name, parameter=None, value=None):
    parameters = {}
    if context.table:
        for row in context.table:
            parameters[row['parameter'].strip()] = eval(row['value'].strip(), {}, {})

    if parameter and value:
        parameters[parameter.strip()] = eval(value.strip(), {}, {})

    context.interpreter.queue(name, **parameters)


@given('I wait {seconds:g} seconds')
@given('I wait {seconds:g} second')
@when('I wait {seconds:g} seconds')
@when('I wait {seconds:g} second')
def wait(context, seconds):
    context.interpreter.clock.time += seconds


@then('state {name} is entered')
def state_is_entered(context, name):
    # Check that state exists
    context.interpreter.statechart.state_for(name)

    test = testing.state_is_entered(context.monitored_trace, name)
    assert test, 'State {} is not entered'.format(name)


@then('state {name} is not entered')
def state_is_not_entered(context, name):
    # Check that state exists
    context.interpreter.statechart.state_for(name)

    test = not testing.state_is_entered(context.monitored_trace, name)
    assert test, 'State {} is entered'.format(name)


@then('state {name} is exited')
def state_is_exited(context, name):
    # Check that state exists
    context.interpreter.statechart.state_for(name)

    test = testing.state_is_exited(context.monitored_trace, name)
    assert test, 'State {} is not exited'.format(name)


@then('state {name} is not exited')
def state_is_not_exited(context, name):
    # Check that state exists
    context.interpreter.statechart.state_for(name)

    test = not testing.state_is_exited(context.monitored_trace, name)
    assert test, 'State {} is exited'.format(name)


@then('state {name} is active')
def state_is_active(context, name):
    # Check that state exists
    context.interpreter.statechart.state_for(name)

    assert name in context.interpreter.configuration, 'State {} is not active'.format(name)


@then('state {name} is not active')
def state_is_not_active(context, name):
    # Check that state exists
    context.interpreter.statechart.state_for(name)

    assert name not in context.interpreter.configuration, 'State {} is active'.format(name)


@then('event {name} is fired')
@then('event {name} is fired with {parameter}={value}')
def event_is_fired(context, name, parameter=None, value=None):
    parameters = {}

    for row in context.table if context.table else []:
        parameters[row['parameter'].strip()] = eval(row['value'].strip(), {}, {})

    if parameter and value:
        parameters[parameter.strip()] = eval(value.strip(), {}, {})

    test = testing.event_is_fired(context.monitored_trace, name, parameters)

    if len(parameters) == 0:
        assert test, 'Event {} is not fired'.format(name)
    else:
        assert test, 'Event {} is not fired with parameters {}'.format(name, parameters)


@then('event {name} is not fired')
def event_is_not_fired(context, name):
    test = not testing.event_is_fired(context.monitored_trace, name)
    assert test, 'Event {} is fired'.format(name)


@then('no event is fired')
def no_event_is_fired(context):
    for macrostep in context.monitored_trace:
        if len(macrostep.sent_events) > 0:
            if len(macrostep.sent_events) > 1:
                assert False, 'Events {} are fired'.format(
                    ', '.join([e.name for e in macrostep.sent_events]))
            else:
                assert False, 'Event {} is fired'.format(macrostep.sent_events[0].name)


@then('variable {variable} equals {value}')
def variable_equals(context, variable, value):
    assert variable in context.interpreter.context, 'Variable {} is not defined'.format(variable)

    current_value = context.interpreter.context[variable]
    expected_value = eval(value, {}, {})
    assert current_value == expected_value, 'Variable {} equals {}, not {}'.format(
        variable, current_value, expected_value)


@then('variable {variable} does not equal {value}')
def variable_does_not_equal(context, variable, value):
    assert variable in context.interpreter.context, 'Variable {} is not defined'.format(variable)

    current_value = context.interpreter.context[variable]
    expected_value = eval(value, {}, {})
    assert current_value != expected_value, 'Variable {} equals {}'.format(variable, current_value)


@then('expression {expression} holds')
def expression_holds(context, expression):
    assert testing.expression_holds(
        context.interpreter, expression), 'Expression {} does not holds'.format(expression)


@then('expression {expression} does not hold')
def expression_does_not_hold(context, expression):
    assert not testing.expression_holds(
        context.interpreter, expression), 'Expression {} holds'.format(expression)


@then('statechart is in a final configuration')
def final_configuration(context):
    assert context.interpreter.final, 'Statechart is not in a final configuration: {}'.format(
        ', '.join(context.interpreter.configuration))


@then('statechart is not in a final configuration')
def not_final_configuration(context):
    assert not context.interpreter.final, 'Statechart is in a final configuration: {}'.format(
        ', '.join(context.interpreter.configuration))



import argparse
import re
import sys

from typing import Dict, List, Tuple, Union
from ..io import import_from_yaml
from ..model import (
    DeepHistoryState, FinalState, Transition, CompoundState,
    OrthogonalState, ShallowHistoryState, Statechart,
    ActionStateMixin, ContractMixin)


__all__ = ['export_to_plantuml']


class PlantUMLExporter:
    def __init__(
            self,
            statechart: Statechart, *,
            based_on: str = None,
            statechart_name: bool = True,
            statechart_description: bool = True,
            statechart_preamble: bool = True,
            state_contracts: bool = True,
            state_action: bool = True,
            transition_contracts: bool = True,
            transition_action: bool = True) -> None:
        self.statechart = statechart
        self.based_on = based_on
        self.statechart_name = statechart_name
        self.statechart_description = statechart_description
        self.statechart_preamble = statechart_preamble
        self.state_contracts = state_contracts
        self.state_action = state_action
        self.transition_contracts = transition_contracts
        self.transition_action = transition_action

        self._based_on_arrows = dict()  # type: Dict[Tuple[str, str], str]
        if self.based_on:
            for line in self.based_on.splitlines():
                matches = re.findall(r'(\[\*\]|[a-zA-Z0-9]+) -([^ ]*)> (\[\*\]|[a-zA-Z0-9]+)', line)
                if matches:
                    self._based_on_arrows[(matches[0][0], matches[0][2])
                                          ] = '-{}>'.format(matches[0][1])

        self._output = []  # type: List[str]
        self._indent = 0

    def arrow(self, source, target):
        # source = None --> initial state
        if not self.based_on:
            return '-->'
        else:
            source = '[*]' if source is None else self.state_id(source)
            target = '[*]' if isinstance(self.statechart.state_for(target),
                                         FinalState) else self.state_id(target)
            return self._based_on_arrows.get((source, target), '-->')

    def indent(self):
        self._indent += 2

    def deindent(self):
        self._indent -= 2

    def output(self, text: str, *, wrap: str = '') -> None:
        lines = text.strip().split('\n')

        for line in lines:
            # Special case for __old__
            line = line.replace('__old__', '~__old__')

            self._output.append(
                '{indent}{wrap}{line}{wrap}'.format(
                    indent=' ' * self._indent,
                    wrap=wrap,
                    line=line
                )
            )

    @staticmethod
    def state_id(name: str) -> str:
        return ''.join(filter(str.isalnum, name))

    def export_statechart(self):
        if self.statechart_name and self.statechart.name:
            self.output('title {}'.format(self.statechart.name))

        if self.statechart_description and self.statechart.description:
            self.output('caption {}'.format(self.statechart.description.replace('\n', '\\n')))

    def export_preamble(self):
        if self.statechart_preamble and self.statechart.preamble:
            self.output('note top of {}'.format(self.state_id(self.statechart.root)))
            self.indent()
            self.output(self.statechart.preamble)
            self.deindent()
            self.output('end note')

    def export_state(self, name: str) -> None:
        state = self.statechart.state_for(name)

        # Special case for final state
        if isinstance(state, FinalState):
            # Find transitions leading to it
            for transition in self.statechart.transitions_to(state.name):
                self.export_transition(transition)
            return

        if isinstance(state, ShallowHistoryState):
            self.output('state "H" as {} {{'.format(self.state_id(name)))
        elif isinstance(state, DeepHistoryState):
            self.output('state "H*" as {} {{'.format(self.state_id(name)))
        else:
            self.output('state "{}" as {} {{'.format(name, self.state_id(name)))

        self.indent()

        # Actions
        has_actions = False

        if self.state_action and isinstance(state, ActionStateMixin):
            # On entry, on exit
            if state.on_entry:
                has_actions = True
                self.output('{} : **entry** / {}'.format(
                    self.state_id(name),
                    state.on_entry.strip().replace('\n', '; ')
                ))
            if state.on_exit:
                has_actions = True
                self.output('{} : **exit** / {}'.format(
                    self.state_id(name),
                    state.on_exit.strip().replace('\n', '; ')
                ))

            # Internal actions
            transitions = [tr for tr in self.statechart.transitions_from(
                name) if tr.internal and tr.action]
            if len(transitions) > 0:
                for transition in transitions:
                    text = []
                    if transition.priority != Transition.DEFAULT_PRIORITY:
                        text.append('{}:'.format(transition.priority))
                    if transition.event:
                        text.append('**{}** '.format(transition.event))
                    if transition.guard:
                        text.append('[{}] '.format(transition.guard))

                    self.output('{} : {}/ {}'.format(
                        self.state_id(name),
                        ''.join(text),
                        transition.action.strip().replace('\n', '; '),
                    ))

        # Contracts
        if isinstance(state, ContractMixin) and self.state_contracts and (
                state.preconditions or state.invariants or state.postconditions):
            if has_actions:
                self.output('{} : '.format(self.state_id(name)))

            for cond in state.preconditions:
                self.output('{} : **pre:** {}'.format(self.state_id(name), cond))
            for cond in state.invariants:
                self.output('{} : **inv:** {}'.format(self.state_id(name), cond))
            for cond in state.postconditions:
                self.output('{} : **post:** {}'.format(self.state_id(name), cond))

        # Transitions
        if isinstance(state, CompoundState) and state.initial:
            self.output('[*] {arrow} {target}'.format(
                arrow=self.arrow(None, state.initial),
                target=self.state_id(state.initial)
            ))

        self.export_transitions(name)

        if isinstance(state, (ShallowHistoryState, DeepHistoryState)):
            self.export_history_memory(state)

        # Nested states
        for i, child in enumerate(self.statechart.children_for(name)):
            if i != 0 and isinstance(state, OrthogonalState):
                self.output('--')
            self.export_state(child)

        self.deindent()
        self.output('}')

    def export_transitions(self, source_name: str) -> None:
        # Transitions (except internal ones)
        transitions = filter(lambda t: not t.internal,
                             self.statechart.transitions_from(source_name))

        for transition in transitions:
            # Do not treat final states here
            if isinstance(self.statechart.state_for(transition.target), FinalState):
                continue
            self.export_transition(transition)

    def export_transition(self, transition: Transition) -> None:
        target = self.statechart.state_for(transition.target)

        if isinstance(target, FinalState):
            target_name = '[*]'
        else:
            target_name = self.state_id(target.name)

        text = []
        if transition.priority != Transition.DEFAULT_PRIORITY:
            text.append('{}:'.format(transition.priority))
        if transition.event:
            text.append(transition.event + ' ')
        if transition.guard:
            text.append('[{}] '.format(transition.guard))
        if transition.action and self.transition_action:
            text.append('/ {}'.format(transition.action.replace('\n', '; ')))

        if self.transition_contracts and (
                transition.preconditions or transition.invariants or transition.postconditions):
            text.append('\\n')

            for cond in transition.preconditions:
                text.append('pre: {}\\n'.format(cond))
            for cond in transition.invariants:
                text.append('inv: {}\n'.format(cond))
            for cond in transition.postconditions:
                text.append('post: {}\n'.format(cond))

        _format = '{source} {arrow} {target} : {text}' if len(
            text) > 0 else '{source} {arrow} {target}'
        self.output(_format.format(
            source=self.state_id(transition.source),
            arrow=self.arrow(transition.source, transition.target),
            target=target_name,
            text=''.join(text),
        ))

    def export_history_memory(self, history_state: Union[ShallowHistoryState, DeepHistoryState]):
        if history_state.memory:
            target = self.statechart.state_for(history_state.memory)

            self.output('{source} {arrow} {target}'.format(
                source=self.state_id(history_state.name),
                arrow=self.arrow(history_state, target),
                target=self.state_id(target.name)
            ))

    def export(self) -> str:
        self.output('@startuml')

        self.export_statechart()
        self.export_state(self.statechart.root)  # type: ignore
        self.export_preamble()

        self.output('@enduml')

        return '\n'.join(self._output)


def export_to_plantuml(
        statechart: Statechart,
        filepath: str = None,
        *,
        based_on: str = None,
        based_on_filepath: str = None,
        statechart_name: bool = True,
        statechart_description: bool = False,
        statechart_preamble: bool = False,
        state_contracts: bool = False,
        state_action: bool = True,
        transition_contracts: bool = False,
        transition_action: bool = True) -> str:
    """
    Export given statechart to plantUML (see http://plantuml/plantuml).
    If a filepath is provided, also save the output to this file.

    Due to the way statecharts are representing, and due to the presence of features that are
    specific to Sismic,the resulting statechart representation does not include all the
    informations.
    For example, final states and history states won't have name, actions and contracts.

    If a previously exported representation for the statechart is provided, either as text
    (based_on parameter) or as a filepath (based_on_filepath parameter), it will attempt to reuse
    the modifications made to the transitions (their direction and length).

    :param statechart: statechart to export
    :param filepath: save output to given filepath, if provided
    :param based_on: existing representation of the statechart in PlantUML
    :param based_on_filepath: filepath to an existing representation of the statechart in PlantUML
    :param statechart_name: include the name of the statechart
    :param statechart_description: include the description of the statechart
    :param statechart_preamble: include the preamble of the statechart
    :param state_contracts: include state contracts
    :param state_action: include state actions (on entry, on exit and internal transitions)
    :param transition_contracts: include transition contracts
    :param transition_action: include actions on transition
    :return: textual representation using plantuml
    """

    if based_on and based_on_filepath:
        raise TypeError(
            'Parameters based_on and based_on_filepath cannot both be provided at the same time.')
    if based_on_filepath:
        with open(based_on_filepath, 'r') as f:
            based_on = f.read()

    exporter = PlantUMLExporter(
        statechart,
        based_on=based_on,
        statechart_name=statechart_name,
        statechart_description=statechart_description,
        statechart_preamble=statechart_preamble,
        state_contracts=state_contracts,
        state_action=state_action,
        transition_contracts=transition_contracts,
        transition_action=transition_action,
    )

    output = exporter.export()

    if filepath:
        with open(filepath, 'w') as f:
            f.write(output)

    return output


def cli(args=None) -> int:
    parser = argparse.ArgumentParser(
        prog='sismic-plantuml',
        description='Command-line utility to export Sismic statecharts to plantUML.\n'
                    'See sismic.io.export_to_plantuml for more informations.'
    )

    parser.add_argument('statechart', metavar='statechart', type=str,
                        help='A YAML file describing a statechart')
    parser.add_argument('--based-on', metavar='based', type=str, default=None,
                        help='A previously exported PlantUML representation for this statechart.')

    parser.add_argument('--show-description', dest='statechart_description',
                        action='store_true', default=False, help='Show statechart description')
    parser.add_argument('--show-preamble', dest='statechart_preamble',
                        action='store_true', default=False, help='Show statechart preamble')
    parser.add_argument('--show-state-contracts', dest='state_contracts',
                        action='store_true', default=False, help='Show state contracts')
    parser.add_argument('--show-transition-contracts', dest='transition_contracts',
                        action='store_false', default=False, help='Show transition contracts')
    parser.add_argument('--hide-state-action', dest='state_action',
                        action='store_false', default=True, help='Hide state action')
    parser.add_argument('--hide-name', dest='statechart_name',
                        action='store_false', default=True, help='Hide statechart name')
    parser.add_argument('--hide-transition-action', dest='transition_action',
                        action='store_false', default=True, help='Hide transition action')

    args, parameters = parser.parse_known_args(args)

    statechart = import_from_yaml(filepath=args.statechart)

    if args.based_on:
        with open(args.based_on, 'r') as f:
            args.based_on = f.read()

    exporter = PlantUMLExporter(
        statechart,
        based_on=args.based_on,
        statechart_name=args.statechart_name,
        statechart_description=args.statechart_description,
        statechart_preamble=args.statechart_preamble,
        state_contracts=args.state_contracts,
        state_action=args.state_action,
        transition_contracts=args.transition_contracts,
        transition_action=args.transition_action,
    )

    print(exporter.export())
    return 0


if __name__ == '__main__':
    sys.exit(cli())



import warnings
from typing import Any

__all__ = ['Event', 'InternalEvent', 'MetaEvent']


class Event:
    """
    An event with a name and (optionally) some data passed as named parameters.

    The list of parameters can be obtained using *dir(event)*. Notice that
    *name* and *data* are reserved names. If a *delay* parameter is provided,
    then this event will be considered as a delayed event (and won't be
    executed until given delay has elapsed).

    When two events are compared, they are considered equal if their names
    and their data are equal.

    :param name: name of the event.
    :param data: additional data passed as named parameters.
    """

    __slots__ = ['name', 'data']

    def __init__(self, name: str, **additional_parameters: Any) -> None:
        self.name = name
        self.data = additional_parameters

    def __eq__(self, other):
        if isinstance(other, Event):
            return (self.name == other.name and self.data == other.data)
        else:
            return NotImplemented

    def __getattr__(self, attr):
        try:
            return self.data[attr]
        except KeyError:
            raise AttributeError('{} has no attribute {}'.format(self, attr))

    def __getstate__(self):
        # For pickle and implicitly for multiprocessing
        return self.name, self.data

    def __setstate__(self, state):
        # For pickle and implicitly for multiprocessing
        self.name, self.data = state

    def __hash__(self):
        return hash(self.name)

    def __dir__(self):
        return ['name'] + list(self.data.keys())

    def __repr__(self):
        if self.data:
            return '{}({!r}, {})'.format(
                self.__class__.__name__, self.name, ', '.join(
                    '{}={!r}'.format(k, v) for k, v in self.data.items()))
        else:
            return '{}({!r})'.format(self.__class__.__name__, self.name)


class InternalEvent(Event):
    """
    Subclass of Event that represents an internal event.
    """
    pass


class DelayedEvent(Event):
    """
    An event that is delayed.

    Deprecated since 1.4.0, use `Event` with a `delay` parameter instead.
    """

    def __init__(self, name: str, delay: float, **additional_parameters: Any) -> None:
        warnings.warn(
            'DelayedEvent is deprecated since 1.4.0, use Event with a delay parameter instead.',
            DeprecationWarning)
        super().__init__(name, delay=delay, **additional_parameters)


class MetaEvent(Event):
    """
    Subclass of Event that represents a MetaEvent, as used in property statecharts.
    """
    pass



import collections
import copy

from types import CodeType
from typing import Any, Dict, List, Optional, Mapping, Iterator

from . import Evaluator
from ..exceptions import CodeEvaluationError
from ..model import Event, InternalEvent, MetaEvent, Transition


__all__ = ['PythonEvaluator']


class FrozenContext(collections.abc.Mapping):
    """
    A shallow copy of a context. The keys of the underlying context are
    exposed as attributes.
    """
    __slots__ = ['__frozencontext']

    def __init__(self, context: Dict) -> None:
        self.__frozencontext = {k: copy.copy(v) for k, v in context.items()}

    def __getattr__(self, item):
        try:
            return self.__frozencontext[item]
        except KeyError:
            raise AttributeError('{} has no attribute {}'.format(self, item))

    def __getstate__(self):
        return self.__frozencontext

    def __setstate__(self, state):
        self.__frozencontext = state

    def __getitem__(self, key):
        return self.__frozencontext[key]

    def __len__(self):
        return len(self.__frozencontext)

    def __iter__(self):
        return iter(self.__frozencontext)


class PythonEvaluator(Evaluator):
    """
    A code evaluator that understands Python.

    This evaluator exposes some additional functions/variables:

    - On both code execution and code evaluation:
        - A *time: float* value that represents the current time exposed by interpreter clock.
        - An *active(name: str) -> bool* Boolean function that takes a state name and return *True*
          if and only if this state is currently active, ie. it is in the active configuration of
          the ``Interpreter`` instance that makes use of this evaluator.
    - On code execution:
        - A *send(name: str, **kwargs) -> None* function that takes an event name and additional
          keyword parameters and raises an internal event with it. Raised events are propagated to
          bound statecharts as external events and to the current statechart as internal event.
          If delay is provided, a delayed event is created.
        - A *notify(name: str, **kwargs) -> None* function that takes an event name and additional
          keyword parameters and raises a meta-event with it. Meta-events are only sent to bound
          property statecharts.
        - If the code is related to a transition, the *event: Event* that fires the transition
          is exposed.
        - A *setdefault(name:str, value: Any) -> Any* function that defines and returns variable
          *name* in the global scope if it is not yet defined.
    - On guard or contract evaluation:
        - If the code is related to a transition, an *event: Optional[Event]* variable is exposed.
          This variable contains the currently considered event, or None.
    - On guard or contract (except preconditions) evaluation:
        - An *after(sec: float) -> bool* Boolean function that returns *True* if and only if the
          source state was entered more than *sec* seconds ago. The time is evaluated according to
          Interpreter's clock.
        - A *idle(sec: float) -> bool* Boolean function that returns *True* if and only if the
          source state did not fire a transition for more than *sec* ago. The time is evaluated
          according to Interpreter's clock.
    - On contract (except preconditions) evaluation:
        - A variable *__old__* that has an attribute *x* for every *x* in the context when either
          the state was entered (if the condition involves a state) or the transition was processed
          (if the condition involves a transition). The value of *__old__.x* is a shallow copy
          of *x* at that time.
    - On contract evaluation:
        - A *sent(name: str) -> bool* function that takes an event name and return True if an
          event with the same name was sent during the current step.
        - A *received(name: str) -> bool* function  that takes an event name and return True if
          an event with the same name is currently processed in this step.

    If an exception occurred while executing or evaluating a piece of code, it is propagated by the
    evaluator.

    :param interpreter: the interpreter that will use this evaluator,
        is expected to be an *Interpreter* instance
    :param initial_context: a dictionary that will be used as *__locals__*
    """

    def __init__(self, interpreter=None, *, initial_context: Mapping[str, Any] = None) -> None:
        super().__init__(interpreter, initial_context=initial_context)

        self._context = {}  # type: Dict[str, Any]
        self._context.update(initial_context if initial_context else {})
        self._interpreter = interpreter

        # Precompiled code
        self._evaluable_code = {}  # type: Dict[str, CodeType]
        self._executable_code = {}  # type: Dict[str, CodeType]

        # Frozen context for __old__
        self._memory = {}  # type: Dict[int, FrozenContext]

    @property
    def context(self) -> Mapping:
        return self._context

    def _setdefault(self, name: str, value: Any) -> Any:
        """
        Define and return variable "name".

        :param name: name of the variable
        :param value: value to use for that variable, if not defined
        :return: value of the variable
        """
        return self._context.setdefault(name, value)

    def _evaluate_code(
            self, code: Optional[str],
            *, additional_context: Mapping[str, Any] = None) -> bool:
        """
        Evaluate given code using Python.

        :param code: code to evaluate
        :param additional_context: an optional additional context
        :return: truth value of *code*
        """
        if code is None:
            return True

        compiled_code = self._evaluable_code.get(code, None)
        if compiled_code is None:
            compiled_code = self._evaluable_code.setdefault(code, compile(code, '<string>', 'eval'))

        exposed_context = {
            'active': lambda s: s in self._interpreter.configuration,
            'time': self._interpreter.time,
        }
        exposed_context.update(additional_context if additional_context is not None else {})

        try:
            return bool(eval(compiled_code, exposed_context, self._context))
        except Exception as e:
            raise CodeEvaluationError('"{}" occurred while evaluating "{}"'.format(e, code)) from e

    def _execute_code(
            self, code: Optional[str],
            *, additional_context: Mapping[str, Any] = None) -> List[Event]:
        """
        Execute given code using Python.

        :param code: code to execute
        :param additional_context: an optional additional context
        :return: a list of sent events
        """
        if code is None:
            return []

        compiled_code = self._executable_code.get(code, None)
        if compiled_code is None:
            compiled_code = self._executable_code.setdefault(
                code, compile(code, '<string>', 'exec'))

        sent_events = []  # type: List[Event]

        exposed_context = {
            'active': lambda name: name in self._interpreter.configuration,
            'time': self._interpreter.time,
            'send': lambda name, **kwargs: sent_events.append(InternalEvent(name, **kwargs)),
            'notify': lambda name, **kwargs: sent_events.append(MetaEvent(name, **kwargs)),
            'setdefault': self._setdefault,
        }
        exposed_context.update(additional_context if additional_context is not None else {})

        try:
            exec(compiled_code, exposed_context, self._context)  # type: ignore
            return sent_events
        except Exception as e:
            raise CodeEvaluationError('"{}" occurred while executing "{}"'.format(e, code)) from e

    def evaluate_guard(self, transition: Transition, event: Optional[Event] = None) -> bool:
        """
        Evaluate the guard for given transition.

        :param transition: the considered transition
        :param event: instance of *Event* if any
        :return: truth value of *code*
        """
        additional_context = {
            'after': (
                lambda seconds: self._interpreter.time - seconds
                >= self._interpreter._entry_time[transition.source]
            ),
            'idle': (
                lambda seconds: self._interpreter.time - seconds
                >= self._interpreter._idle_time[transition.source]
            ),
            'event': event,
        }
        return self._evaluate_code(
            getattr(transition, 'guard', None),
            additional_context=additional_context)

    def evaluate_preconditions(self, obj, event: Optional[Event] = None) -> Iterator[str]:
        """
        Evaluate the preconditions for given object (either a *StateMixin* or a
        *Transition*) and return a list of conditions that are not satisfied.

        :param obj: the considered state or transition
        :param event: an optional *Event* instance, if any
        :return: list of unsatisfied conditions
        """
        additional_context = {
            'received': lambda name: name == getattr(event, 'name', None),
            'sent': lambda name: name in [e.name for e in self._interpreter._sent_events],
            'event': event,
        }

        # Deal with __old__ in contracts, only required if there is an invariant or a postcondition
        if len(getattr(obj, 'invariants', [])) > 0 or len(getattr(obj, 'postconditions', [])) > 0:
            self._memory[id(obj)] = FrozenContext(self._context)

        return filter(
            lambda c: not self._evaluate_code(c, additional_context=additional_context),
            getattr(obj, 'preconditions', [])
        )

    def evaluate_invariants(self, obj, event: Optional[Event] = None) -> Iterator[str]:
        """
        Evaluate the invariants for given object (either a *StateMixin* or a
        *Transition*) and return a list of conditions that are not satisfied.

        :param obj: the considered state or transition
        :param event: an optional *Event* instance, if any
        :return: list of unsatisfied conditions
        """
        state_name = obj.source if isinstance(obj, Transition) else obj.name

        additional_context = {
            '__old__': self._memory.get(
                id(obj),
                None),
            'after': (
                lambda seconds: self._interpreter.time - seconds
                >= self._interpreter._entry_time[state_name]
            ),
            'idle': (
                lambda seconds: self._interpreter.time - seconds
                >= self._interpreter._idle_time[state_name]
            ),
            'received': lambda name: name == getattr(
                event,
                'name',
                None),
            'sent': lambda name: name in [
                e.name for e in self._interpreter._sent_events],
            'event': event,
        }

        return filter(
            lambda c: not self._evaluate_code(c, additional_context=additional_context),
            getattr(obj, 'invariants', [])
        )

    def evaluate_postconditions(self, obj, event: Optional[Event] = None) -> Iterator[str]:
        """
        Evaluate the postconditions for given object (either a *StateMixin* or a
        *Transition*) and return a list of conditions that are not satisfied.

        :param obj: the considered state or transition
        :param event: an optional *Event* instance, if any
        :return: list of unsatisfied conditions
        """
        state_name = obj.source if isinstance(obj, Transition) else obj.name

        additional_context = {
            '__old__': self._memory.get(
                id(obj),
                None),
            'after': (
                lambda seconds: self._interpreter.time - seconds
                >= self._interpreter._entry_time[state_name]
            ),
            'idle': (
                lambda seconds: self._interpreter.time - seconds
                >= self._interpreter._idle_time[state_name]
            ),
            'received': lambda name: name == getattr(
                event,
                'name',
                None),
            'sent': lambda name: name in [
                e.name for e in self._interpreter._sent_events],
            'event': event,
        }

        return filter(
            lambda c: not self._evaluate_code(c, additional_context=additional_context),
            getattr(obj, 'postconditions', [])
        )

    def __getstate__(self):
        attributes = self.__dict__.copy()
        attributes['_executable_code'] = dict()  # Code fragment cannot be pickled
        attributes['_evaluable_code'] = dict()  # Code fragment cannot be pickled
        return attributes



import pytest

from sismic.model import Statechart
from sismic.exceptions import StatechartError
from sismic.io import import_from_yaml, export_to_yaml, export_to_plantuml
from sismic.io.plantuml import cli


def compare_statecharts(s1, s2):
    assert s1.name == s2.name
    assert s1.description == s2.description
    assert s2.preamble == s2.preamble

    assert set(s1.states) == set(s2.states)
    assert set(s1.transitions) == set(s2.transitions)

    for state in s1.states:
        assert s1.parent_for(state) == s2.parent_for(state)
        assert set(s1.children_for(state)) == set(s2.children_for(state))


@pytest.mark.parametrize('data', [1, -1, 1.0, 'yes', 'True', 'no', '', [], [1, 2], {}, {1: 1}])
def test_yaml_parser_types_handling(data):
    yaml = ('statechart:'
            '\n  name: ' + str(data) +
            '\n  preamble: Nothing'
            '\n  root state:'
            '\n    name: s1')
    item = import_from_yaml(yaml).name
    assert isinstance(item, str)


def test_import_from_yaml_args():
    with pytest.raises(TypeError):
        import_from_yaml()
    with pytest.raises(TypeError):
        import_from_yaml('A', filepath='B')


class TestImportFromYaml:
    def test_import_example_from_tests(self, example_from_tests):
        assert isinstance(example_from_tests, Statechart)

    def test_import_example_from_docs(self, example_from_docs):
        assert isinstance(example_from_docs, Statechart)

    def test_transitions_to_unknown_state(self):
        yaml = """
        statechart:
          name: test
          root state:
            name: root
            initial: s1
            states:
              - name: s1
                transitions:
                  - target: s2
        """
        with pytest.raises(StatechartError) as e:
            import_from_yaml(yaml)
        assert 'Unknown target state' in str(e.value)

    def test_history_not_in_compound(self):
        yaml = """
        statechart:
          name: test
          root state:
            name: root
            initial: s1
            states:
              - name: s1
                parallel states:
                 - name: s2
                   type: shallow history
        """
        with pytest.raises(StatechartError) as e:
            import_from_yaml(yaml)
        assert 'cannot be used as a parent for' in str(e.value)

    def test_declare_both_states_and_parallel_states(self):
        yaml = """
        statechart:
          name: test
          root state:
            name: root
            initial: s1
            states:
              - name: s1
            parallel states:
              - name: s2
        """

        with pytest.raises(StatechartError) as e:
            import_from_yaml(yaml)
        assert 'root cannot declare both a "states" and a "parallel states" property' in str(e.value)


class TestExportToYaml:
    def test_export_example_from_tests(self, example_from_tests):
        assert len(export_to_yaml(example_from_tests)) > 0

    def test_export_example_from_docs(self, example_from_docs):
        assert len(export_to_yaml(example_from_docs)) > 0

    def test_validity_for_example_from_tests(self, example_from_tests):
        assert import_from_yaml(export_to_yaml(example_from_tests)).validate()

    def test_validity_for_example_from_docs(self, example_from_docs):
        assert import_from_yaml(export_to_yaml(example_from_docs)).validate()

    def test_identity_for_example_from_tests(self, example_from_tests):
        compare_statecharts(example_from_tests, import_from_yaml(export_to_yaml(example_from_tests)))

    def test_identity_for_example_from_docs(self, example_from_docs):
        compare_statecharts(example_from_docs, import_from_yaml(export_to_yaml(example_from_docs)))


class TestExportToPlantUML:
    def test_export_example_from_tests(self, example_from_tests):
        export = export_to_plantuml(
            example_from_tests,
            statechart_name=True,
            statechart_description=True,
            statechart_preamble=True,
            state_contracts=True,
            state_action=True,
            transition_contracts=True,
            transition_action=True
        )
        assert len(export) > 0

    def test_export_example_from_docs(self, example_from_docs):
        export = export_to_plantuml(
            example_from_docs,
            statechart_name=True,
            statechart_description=True,
            statechart_preamble=True,
            state_contracts=True,
            state_action=True,
            transition_contracts=True,
            transition_action=True
        )
        assert len(export) > 0

    def test_export_based_on_filepath(self, elevator):
        filepath = 'docs/examples/elevator/elevator.plantuml'
        statechart = elevator.statechart
        with open(filepath, 'r') as f:
            p1 = f.read()

        assert p1 != export_to_plantuml(statechart)
        assert p1 == export_to_plantuml(statechart, based_on=p1)
        assert p1 == export_to_plantuml(statechart, based_on_filepath=filepath)


    def test_cli(self, capsys):
        filepath = 'docs/examples/elevator/elevator.yaml'
        statechart = import_from_yaml(filepath=filepath)

        # Check default parameters
        cli([filepath])
        out, _ = capsys.readouterr()
        assert export_to_plantuml(statechart) == out.strip()

        # Check all parameters
        cli([filepath, '--based-on', 'docs/examples/elevator/elevator.plantuml', '--show-description', '--show-preamble', '--show-state-contracts', '--show-transition-contracts', '--hide-state-action', '--hide-name', '--hide-transition-action'])
        out, _ = capsys.readouterr()
        export = export_to_plantuml(statechart, based_on_filepath='docs/examples/elevator/elevator.plantuml', statechart_description=True, statechart_preamble=True, state_contracts=True, transition_contracts=True, state_action=False, statechart_name=False, transition_action=False)
        assert export == out.strip()
        



import pytest

import os

from sismic.exceptions import StatechartError
from sismic.interpreter import Event
from sismic.bdd import steps, execute_bdd
from sismic.bdd.__main__ import cli
from sismic.io import import_from_yaml


def test_elevator(elevator):
    assert 0 == execute_bdd(
        elevator.statechart,
        [os.path.join('docs', 'examples', 'elevator', 'elevator.feature')]
    )


class TestMicrowave:
    @pytest.fixture
    def property_statecharts(self):
        statecharts = []
        for filename in ['heating_off_property', 'heating_on_property', 'heating_property']:
            with open(os.path.join('docs', 'examples', 'microwave', filename+'.yaml')) as f:
                statecharts.append(import_from_yaml(f))
        return statecharts

    def test_microwave(self, microwave):
        assert 0 == execute_bdd(
            microwave.statechart,
            [os.path.join('docs', 'examples', 'microwave', 'heating.feature')]
        )

    def test_microwave_with_properties(self, microwave, property_statecharts):
        assert 0 == execute_bdd(
            microwave.statechart,
            [os.path.join('docs', 'examples', 'microwave', 'heating.feature')],
            property_statecharts=property_statecharts
        )

    def test_microwave_with_steps(self, microwave):
        features = ['cooking_human', 'lighting_human', 'safety_human']

        assert 0 == execute_bdd(
            microwave.statechart,
            [os.path.join('docs', 'examples', 'microwave', f+'.feature') for f in features],
            step_filepaths=[os.path.join('docs', 'examples', 'microwave', 'steps.py')],
        )

    def test_microwave_with_steps_and_properties(self, microwave, property_statecharts):
        features = ['cooking_human', 'lighting_human', 'safety_human']

        assert 0 == execute_bdd(
            microwave.statechart,
            [os.path.join('docs', 'examples', 'microwave', f+'.feature') for f in features],
            step_filepaths=[os.path.join('docs', 'examples', 'microwave', 'steps.py')],
            property_statecharts=property_statecharts
        )


def test_cli():
    assert 0 == cli([
        'docs/examples/microwave/microwave.yaml',
        '--features', 'docs/examples/microwave/heating.feature', 'docs/examples/microwave/cooking_human.feature', 'docs/examples/microwave/lighting_human.feature', 'docs/examples/microwave/safety_human.feature',
        '--steps', 'docs/examples/microwave/steps.py',
        '--properties', 'docs/examples/microwave/heating_on_property.yaml', 'docs/examples/microwave/heating_off_property.yaml'
    ])


class TestSteps:
    @pytest.fixture
    def context(self, mocker):
        context = mocker.MagicMock(name='context')

        context.interpreter = mocker.MagicMock(name='interpreter')
        context.interpreter.queue = mocker.MagicMock(name='queue')
        context.interpreter.context = {'x': 1}

        def state_for(name):
            if name == 'unknown state':
                raise StatechartError()
            else:
                return mocker.DEFAULT

        context.interpreter.statechart.state_for = mocker.MagicMock(side_effect=state_for)
        context.execute_steps = mocker.MagicMock(name='execute_steps')
        context.monitored_trace = []

        return context

    @pytest.fixture
    def trace(self, mocker):
        macrostep = mocker.MagicMock()
        macrostep.entered_states = ['a', 'b', 'c']
        macrostep.exited_states = ['x', 'y', 'z']
        macrostep.sent_events = [Event('e1'), Event('e2'), Event('e3')]

        return [macrostep]

    def test_do_nothing(self, context):
        steps.do_nothing(context)
        context.interpreter.queue.assert_not_called()

    def test_repeat_step(self, context):
        steps.repeat_step(context, 'blabla', 3)
        assert context.execute_steps.call_count == 3
        context.execute_steps.assert_called_with('Given blabla')

    def test_send_event(self, context):
        steps.send_event(context, 'event_name')
        context.interpreter.queue.assert_called_with('event_name')

    def test_send_event_with_parameter(self, context):
        steps.send_event(context, 'event_name', 'x', '1')
        context.interpreter.queue.assert_called_with('event_name', x=1)

    def test_wait(self, context):
        context.interpreter.clock.time = 0
        steps.wait(context, 3)
        assert context.interpreter.clock.time == 3
        
        steps.wait(context, 6)
        assert context.interpreter.clock.time == 9

    def test_state_is_entered(self, context, trace):
        context.monitored_trace = []
        with pytest.raises(AssertionError):
            steps.state_is_entered(context, 'state')

        context.monitored_trace = trace
        steps.state_is_entered(context, 'a')
        steps.state_is_entered(context, 'b')
        steps.state_is_entered(context, 'c')
        with pytest.raises(AssertionError):
            steps.state_is_entered(context, 'state')
        with pytest.raises(StatechartError):
            steps.state_is_entered(context, 'unknown state')

    def test_state_is_not_entered(self, context, trace):
        context.monitored_trace = []
        steps.state_is_not_entered(context, 'state')

        context.monitored_trace = trace
        with pytest.raises(AssertionError):
            steps.state_is_not_entered(context, 'a')
        steps.state_is_not_entered(context, 'state')
        with pytest.raises(StatechartError):
            steps.state_is_entered(context, 'unknown state')

    def test_state_is_exited(self, context, trace):
        context.monitored_trace = []
        with pytest.raises(AssertionError):
            steps.state_is_exited(context, 'state')

        context.monitored_trace = trace
        steps.state_is_exited(context, 'x')
        steps.state_is_exited(context, 'y')
        steps.state_is_exited(context, 'z')
        with pytest.raises(AssertionError):
            steps.state_is_entered(context, 'state')
        with pytest.raises(StatechartError):
            steps.state_is_entered(context, 'unknown state')

    def test_state_is_not_exited(self, context, trace):
        context.monitored_trace = []
        steps.state_is_not_exited(context, 'state')

        context.monitored_trace = trace
        with pytest.raises(AssertionError):
            steps.state_is_not_exited(context, 'x')
        steps.state_is_not_exited(context, 'state')
        with pytest.raises(StatechartError):
            steps.state_is_entered(context, 'unknown state')

    def test_state_is_active(self, context):
        context.interpreter.configuration = []
        with pytest.raises(AssertionError):
            steps.state_is_active(context, 'state')

        context.interpreter.configuration = ['a', 'b', 'c']
        steps.state_is_active(context, 'a')
        steps.state_is_active(context, 'b')
        steps.state_is_active(context, 'c')
        with pytest.raises(AssertionError):
            steps.state_is_active(context, 'state')
        with pytest.raises(StatechartError):
            steps.state_is_entered(context, 'unknown state')

    def test_state_is_not_active(self, context):
        context.interpreter.configuration = []

        steps.state_is_not_active(context, 'state')

        context.interpreter.configuration = ['a', 'b', 'c']
        with pytest.raises(AssertionError):
            steps.state_is_not_active(context, 'a')

        steps.state_is_not_active(context, 'state')
        with pytest.raises(StatechartError):
            steps.state_is_entered(context, 'unknown state')

    def test_event_is_fired(self, context, trace):
        context.monitored_trace = []
        with pytest.raises(AssertionError):
            steps.event_is_fired(context, 'event')

        context.monitored_trace = trace
        steps.event_is_fired(context, 'e1')
        steps.event_is_fired(context, 'e2')
        steps.event_is_fired(context, 'e3')

        with pytest.raises(AssertionError):
            steps.event_is_fired(context, 'event')

    def test_event_is_fired_with_parameter(self, context, trace):
        context.monitored_trace = trace
        context.monitored_trace[0].sent_events = [Event('e1', x=1)]

        steps.event_is_fired(context, 'e1')

        with pytest.raises(AssertionError):
            steps.event_is_fired(context, 'event')

        with pytest.raises(AssertionError):
            steps.event_is_fired(context, 'e1', 'x', '2')

        steps.event_is_fired(context, 'e1', 'x', '1')

    def test_event_is_not_fired(self, context, trace):
        context.monitored_trace = []

        steps.event_is_not_fired(context, 'event')

        context.monitored_trace = trace
        with pytest.raises(AssertionError):
            steps.event_is_not_fired(context, 'e1')

        steps.event_is_not_fired(context, 'event')

    def test_no_event_is_fired(self, context, trace):
        context.monitored_trace = []
        steps.no_event_is_fired(context)

        context.monitored_trace = trace
        with pytest.raises(AssertionError):
            steps.no_event_is_fired(context)

    def test_variable_equals(self, context):
        steps.variable_equals(context, 'x', '1')

        with pytest.raises(AssertionError):
            steps.variable_equals(context, 'y', '1')

        with pytest.raises(AssertionError):
            steps.variable_equals(context, 'x', '2')

    def test_variable_does_not_equal(self, context):
        steps.variable_does_not_equal(context, 'x', '2')

        with pytest.raises(AssertionError):
            steps.variable_does_not_equal(context, 'x', '1')

        with pytest.raises(AssertionError):
            steps.variable_does_not_equal(context, 'y', '1')

    def test_final_configuration(self, context):
        context.interpreter.final = True
        steps.final_configuration(context)

        context.interpreter.final = False
        with pytest.raises(AssertionError):
            steps.final_configuration(context)

    def test_not_final_configuration(self, context):
        context.interpreter.final = False
        steps.not_final_configuration(context)

        context.interpreter.final = True
        with pytest.raises(AssertionError):
            steps.not_final_configuration(context)



from typing import List, Optional

from .elements import Transition
from .events import Event

__all__ = ['MicroStep', 'MacroStep']


class MicroStep:
    """
    Create a micro step.

    A step consider *event*, takes a *transition* and results in a list
    of *entered_states* and a list of *exited_states*.
    Order in the two lists is REALLY important!

    :param event: Event or None in case of eventless transition
    :param transition: a *Transition* or None if no processed transition
    :param entered_states: possibly empty list of entered states
    :param exited_states: possibly empty list of exited states
    :param sent_events: a possibly empty list of events that are sent during the step
    """

    __slots__ = ['event', 'transition', 'entered_states', 'exited_states', 'sent_events']

    def __init__(self, event: Event = None, transition: Transition = None,
                 entered_states: List[str] = None, exited_states: List[str] = None,
                 sent_events: List[Event] = None) -> None:
        self.event = event
        self.transition = transition
        self.entered_states = entered_states if entered_states else []  # type: List[str]
        self.exited_states = exited_states if exited_states else []  # type: List[str]
        self.sent_events = sent_events if sent_events else []  # type: List[Event]

    def __repr__(self):
        params = []
        if self.event:
            params.append('event={!r}'.format(self.event))
        if self.transition:
            params.append('transition={!r}'.format(self.transition))
        if self.entered_states:
            params.append('entered_states={!r}'.format(self.entered_states))
        if self.exited_states:
            params.append('exited_states={!r}'.format(self.exited_states))
        if self.sent_events:
            params.append('sent_events={!r}'.format(self.sent_events))
        return '{}({})'.format(self.__class__.__name__, ', '.join(params))


class MacroStep:
    """
    A macro step is a list of micro steps.

    :param time: the time at which this step was executed
    :param steps: a list of *MicroStep* instances
    """

    def __init__(self, time: float, steps: List[MicroStep]) -> None:
        self._time = time
        self._steps = steps

    __slots__ = ['_time', '_steps']

    @property
    def steps(self) -> List[MicroStep]:
        """
        List of micro steps
        """
        return self._steps

    @property
    def time(self) -> float:
        """
        Time at which this step was executed.
        """
        return self._time

    @property
    def event(self) -> Optional[Event]:
        """
        Event (or *None*) that was consumed.
        """
        for step in self._steps:
            if step.event:
                return step.event
        return None

    @property
    def transitions(self) -> List[Transition]:
        """
        A (possibly empty) list of transitions that were triggered.
        """
        return [step.transition for step in self._steps if step.transition]

    @property
    def entered_states(self) -> List[str]:
        """
        List of the states names that were entered.
        """
        states = []  # type: List[str]
        for step in self._steps:
            states += step.entered_states
        return states

    @property
    def exited_states(self) -> List[str]:
        """
        List of the states names that were exited.
        """
        states = []  # type: List[str]
        for step in self._steps:
            states += step.exited_states
        return states

    @property
    def sent_events(self) -> List[Event]:
        """
        List of events that were sent during this step.
        """
        events = []
        for step in self._steps:
            for event in step.sent_events:
                events.append(event)
        return events

    def __repr__(self):
        return '{}({!r}, {!r})'.format(self.__class__.__name__, self.time, self._steps)

    def __str__(self):
        return 'Step@{}({}, {}, >{}, <{})'.format(round(self.time, 3), self.event, self.transitions,
                                                  self.entered_states, self.exited_states)



import threading
import time
import warnings

from collections import Counter
from functools import wraps
from typing import Any, Callable, List, Mapping

from .interpreter import Interpreter
from .model import MacroStep

__all__ = ['log_trace', 'run_in_background', 'coverage_from_trace']


def log_trace(interpreter: Interpreter) -> List[MacroStep]:
    """
    Return a list that will be populated by each value returned by the *execute_once* method
    of given interpreter.

    :param interpreter: an *Interpreter* instance
    :return: a list of *MacroStep*
    """
    func = interpreter.execute_once
    trace = []

    @wraps(func)
    def new_func():
        step = func()
        if step:
            trace.append(step)
        return step

    interpreter.execute_once = new_func  # type: ignore
    return trace


def coverage_from_trace(trace: List[MacroStep]) -> Mapping[str, Counter]:
    """
    Given a list of macro steps considered as the trace of a statechart execution, return *Counter*
    objects that counts the states that were entered, the states that were exited and the
    transitions that were processed.

    :param trace: A list of macro steps
    :return: A dict whose keys are "entered states", "exited states" and "processed transitions"
    and whose values are Counter object.
    """
    entered_states = []
    exited_states = []
    processed_transitions = []

    for macrostep in trace:
        for microstep in macrostep.steps:
            entered_states.extend(microstep.entered_states)
            exited_states.extend(microstep.exited_states)
            if microstep.transition:
                processed_transitions.append(microstep.transition)

    return {
        'entered states': Counter(entered_states),
        'exited states': Counter(exited_states),
        'processed transitions': Counter(processed_transitions)
    }


def run_in_background(interpreter: Interpreter,
                      delay: float = 0.05,
                      callback: Callable[[List[MacroStep]], Any] = None) -> threading.Thread:
    """
    Run given interpreter in background.

    The interpreter is ran until it reaches a final configuration.
    You can manually stop the thread using the added *stop* of the returned Thread object.
    This is for convenience only and should be avoided, because a call to *stop* puts the
    interpreter in an empty (and thus final) configuration, without properly leaving the
    active states.

    :param interpreter: an interpreter
    :param delay: delay between each call to *execute()*
    :param callback: a function that accepts the result of *execute*.
    :return: started thread (instance of *threading.Thread*)
    :deprecated: since 1.3.0, use runner.AsyncRunner instead.
    """
    warnings.warn('Deprecated since 1.3.0. Use runner.AsyncRunner instead.', DeprecationWarning)

    def _task():
        while not interpreter.final:
            steps = interpreter.execute()
            if callback:
                callback(steps)
            time.sleep(delay)
    thread = threading.Thread(target=_task)

    def stop_thread():
        interpreter._configuration = set()

    thread.stop = stop_thread  # type: ignore

    thread.start()
    return thread



import time
import threading

from typing import List

from ..interpreter import Interpreter
from ..model import MacroStep


__all__ = ['AsyncRunner']


class AsyncRunner:
    """
    An asynchronous runner that repeatedly execute given interpreter.

    The runner tries to call its `execute` method every `interval` seconds, assuming
    that a call to that method takes less time than `interval`.
    If not, subsequent call is queued and will occur as soon as possible with
    no delay. The runner stops as soon as the underlying interpreter reaches
    a final configuration.

    The execution must be started with the `start` method, and can be (definitively)
    stopped with the `stop` method. An execution can be temporarily suspended
    using the `pause` and `unpause` methods. A call to `wait` blocks until
    the statechart reaches a final configuration.

    The current state of a runner can be obtained using its `running` and
    `paused` properties.

    While this runner can be used "as is", it is designed to be subclassed and
    as such, proposes several hooks to control the execution and additional
    behaviours:

     - before_run: called (only once!) when the runner is started. By default, does nothing.
     - after_run: called (only once!) when the interpreter reaches a final configuration.
       configuration of the underlying interpreter is reached. By default, does nothing.
     - execute: called at each step of the run. By default, calls the `execute_once`
       method of the underlying interpreter and returns a *list* of macro steps.
     - before_execute: called right before the call to `execute()`. By default, does nothing.
     - after_execute: called right after the call to `execute()` with the returned value
       of `execute()`. By default, does nothing.

    By default, this runner calls the interpreter's `execute_once` method only once per cycle
    (meaning at least one macro step is processed during each cycle). If `execute_all` is
    set to True, then `execute_once` is repeatedly called until no macro step can be
    processed in the current cycle.

    :param interpreter: interpreter instance to run.
    :param interval: interval between two calls to `execute`
    :param execute_all: Repeatedly call interpreter's `execute_once` method at each step.
    """

    def __init__(self, interpreter: Interpreter, interval: float = 0.1, execute_all=False) -> None:
        self._unpaused = threading.Event()
        self._stop = threading.Event()

        self.interpreter = interpreter
        self.interval = interval
        self._execute_all = execute_all
        self._thread = threading.Thread(target=self._run)

    @property
    def running(self):
        """
        Holds if execution is currently running (even if it's paused).
        """
        return self._thread.is_alive()

    @property
    def paused(self):
        """
        Holds if execution is running but paused.
        """
        return self.running and not self._unpaused.is_set()

    def start(self):
        """
        Start the execution.
        """
        if self._stop.is_set():
            raise RuntimeError('Cannot restart a stopped runner.')
        elif self._thread.is_alive():
            raise RuntimeError('Runner is already started')
        else:
            self._unpaused.set()
            self._thread.start()

    def stop(self):
        """
        Stop the execution.
        """
        self._stop.set()
        self._unpaused.set()
        self.wait()

    def pause(self):
        """
        Pause the execution.
        """
        self._unpaused.clear()

    def unpause(self):
        """
        Unpause the execution.
        """
        self._unpaused.set()

    def wait(self):
        """
        Wait for the execution to finish.
        """
        if self._thread.is_alive():
            self._thread.join()

    def execute(self) -> List[MacroStep]:
        """
        Called each time the interpreter has to be executed.
        """
        steps = []
        step = self.interpreter.execute_once()

        while step:
            steps.append(step)
            step = self.interpreter.execute_once()

            if not self._execute_all:
                break

        return steps

    def before_execute(self):
        """
        Called before each call to `execute()`.
        """
        pass

    def after_execute(self, steps: List[MacroStep]):
        """
        Called after each call to self.execute().
        Receives the return value of self.execute().

        :param steps: List of macrosteps returned by self.execute()
        """
        pass

    def before_run(self):
        """
        Called before running the execution.
        """
        pass

    def after_run(self):
        """
        Called after a final configuration is reached.
        """
        pass

    def _run(self):
        self.before_run()
        self._unpaused.wait()

        while not self.interpreter.final and not self._stop.is_set():
            starttime = time.time()
            self.before_execute()
            r = self.execute()
            self.after_execute(r)

            elapsed = time.time() - starttime
            time.sleep(max(0, self.interval - elapsed))
            self._unpaused.wait()

        # Ensure that self._stop is set if self.interpreter.final holds
        self._stop.set()

        self.after_run()

    def __del__(self):
        self.stop()



from abc import ABCMeta
from typing import List

__all__ = ['ContractMixin', 'StateMixin', 'ActionStateMixin', 'TransitionStateMixin',
           'CompositeStateMixin', 'HistoryStateMixin', 'BasicState', 'CompoundState',
           'OrthogonalState', 'ShallowHistoryState', 'DeepHistoryState', 'FinalState', 'Transition']


class ContractMixin(metaclass=ABCMeta):
    """
    Mixin with a contract: preconditions, postconditions and invariants.
    """

    def __init__(self) -> None:
        self.preconditions = []  # type: List[str]
        self.postconditions = []  # type: List[str]
        self.invariants = []  # type: List[str]

    def __eq__(self, other):
        if isinstance(other, ContractMixin):
            return (
                self.preconditions == other.preconditions
                and self.postconditions == other.postconditions
                and self.invariants == other.invariants
            )
        else:
            return NotImplemented


class StateMixin(metaclass=ABCMeta):
    """
    State element with a name.

    :param name: name of the state
    """

    def __init__(self, name: str) -> None:
        self._name = name

    @property
    def name(self):
        return self._name

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, self.name)

    def __eq__(self, other):
        if isinstance(other, StateMixin):
            return self.name == other.name
        else:
            return NotImplemented

    def __hash__(self):
        return hash(self.name)


class ActionStateMixin(metaclass=ABCMeta):
    """
    State that can define actions on entry and on exit.

    :param on_entry: code to execute when state is entered
    :param on_exit: code to execute when state is exited
    """

    def __init__(self, on_entry: str = None, on_exit: str = None) -> None:
        self.on_entry = on_entry
        self.on_exit = on_exit

    def __eq__(self, other):
        if isinstance(other, ActionStateMixin):
            return self.on_entry == other.on_exit and self.on_exit == other.on_exit
        else:
            return NotImplemented


class TransitionStateMixin(metaclass=ABCMeta):
    """
    A simple state can host transitions
    """

    def __eq__(self, other):
        return isinstance(other, TransitionStateMixin)


class CompositeStateMixin(metaclass=ABCMeta):
    """
    Composite state can have children states.
    """

    def __eq__(self, other):
        return isinstance(other, CompositeStateMixin)


class HistoryStateMixin(metaclass=ABCMeta):
    """
    History state has a memory that can be resumed.

    :param memory: name of the initial state
    """

    def __init__(self, memory: str = None) -> None:
        self.memory = memory

    def __eq__(self, other):
        if isinstance(other, HistoryStateMixin):
            return self.memory == other.memory
        else:
            return NotImplemented


class BasicState(ContractMixin, StateMixin, ActionStateMixin, TransitionStateMixin):
    """
    A basic state, with a name, transitions, actions, etc. but no child state.

    :param name: name of this state
    :param on_entry: code to execute when state is entered
    :param on_exit: code to execute when state is exited
    """

    def __init__(self, name: str, on_entry: str = None, on_exit: str = None) -> None:
        ContractMixin.__init__(self)
        StateMixin.__init__(self, name)
        ActionStateMixin.__init__(self, on_entry, on_exit)
        TransitionStateMixin.__init__(self)

    def __eq__(self, other):
        if isinstance(other, BasicState):
            return (
                ContractMixin.__eq__(self, other)
                and StateMixin.__eq__(self, other)
                and ActionStateMixin.__eq__(self, other)
                and TransitionStateMixin.__eq__(self, other)
            )
        else:
            return NotImplemented


class CompoundState(
        ContractMixin, StateMixin, ActionStateMixin, TransitionStateMixin, CompositeStateMixin):
    """
    Compound states must have children states.

    :param name: name of this state
    :param initial: name of the initial state
    :param on_entry: code to execute when state is entered
    :param on_exit: code to execute when state is exited
    """

    def __init__(self, name: str, initial: str = None,
                 on_entry: str = None, on_exit: str = None) -> None:
        ContractMixin.__init__(self)
        StateMixin.__init__(self, name)
        ActionStateMixin.__init__(self, on_entry, on_exit)
        TransitionStateMixin.__init__(self)
        CompositeStateMixin.__init__(self)
        self.initial = initial

    def __eq__(self, other):
        if isinstance(other, CompoundState):
            return (
                ContractMixin.__eq__(self, other)
                and StateMixin.__eq__(self, other)
                and ActionStateMixin.__eq__(self, other)
                and TransitionStateMixin.__eq__(self, other)
                and CompositeStateMixin.__eq__(self, other)
            )
        else:
            return NotImplemented


class OrthogonalState(
        ContractMixin, StateMixin, ActionStateMixin, TransitionStateMixin, CompositeStateMixin):
    """
    Orthogonal states run their children simultaneously.

    :param name: name of this state
    :param on_entry: code to execute when state is entered
    :param on_exit: code to execute when state is exited
    """

    def __init__(self, name: str, on_entry: str = None, on_exit: str = None) -> None:
        ContractMixin.__init__(self)
        StateMixin.__init__(self, name)
        ActionStateMixin.__init__(self, on_entry, on_exit)
        TransitionStateMixin.__init__(self)
        CompositeStateMixin.__init__(self)

    def __eq__(self, other):
        if isinstance(other, OrthogonalState):
            return (
                ContractMixin.__eq__(self, other)
                and StateMixin.__eq__(self, other)
                and ActionStateMixin.__eq__(self, other)
                and TransitionStateMixin.__eq__(self, other)
                and CompositeStateMixin.__eq__(self, other)
            )
        else:
            return NotImplemented


class ShallowHistoryState(ContractMixin, StateMixin, ActionStateMixin, HistoryStateMixin):
    """
    A shallow history state resumes the execution of its parent.
    It activates the latest visited state of its parent.

    :param name: name of this state
    :param on_entry: code to execute when state is entered
    :param on_exit: code to execute when state is exited
    :param memory: name of the initial state
    """

    def __init__(self, name: str, on_entry: str = None,
                 on_exit: str = None, memory: str = None) -> None:
        ContractMixin.__init__(self)
        StateMixin.__init__(self, name)
        ActionStateMixin.__init__(self, on_entry, on_exit)
        HistoryStateMixin.__init__(self, memory)

    def __eq__(self, other):
        if isinstance(other, ShallowHistoryState):
            return (
                ContractMixin.__eq__(self, other)
                and StateMixin.__eq__(self, other)
                and ActionStateMixin.__eq__(self, other)
                and HistoryStateMixin.__eq__(self, other)
            )
        else:
            return NotImplemented


class DeepHistoryState(ContractMixin, StateMixin, ActionStateMixin, HistoryStateMixin):
    """
    A deep history state resumes the execution of its parent, and of every nested
    active states in its parent.

    :param name: name of this state
    :param on_entry: code to execute when state is entered
    :param on_exit: code to execute when state is exited
    :param memory: name of the initial state
    """

    def __init__(self, name: str, on_entry: str = None,
                 on_exit: str = None, memory: str = None) -> None:
        ContractMixin.__init__(self)
        StateMixin.__init__(self, name)
        ActionStateMixin.__init__(self, on_entry, on_exit)
        HistoryStateMixin.__init__(self, memory)

    def __eq__(self, other):
        if isinstance(other, DeepHistoryState):
            return (
                ContractMixin.__eq__(self, other)
                and StateMixin.__eq__(self, other)
                and ActionStateMixin.__eq__(self, other)
                and HistoryStateMixin.__eq__(self, other)
            )
        else:
            return NotImplemented


class FinalState(ContractMixin, StateMixin, ActionStateMixin):
    """
    Final state has NO transition and is used to detect state machine termination.

    :param name: name of this state
    :param on_entry: code to execute when state is entered
    :param on_exit: code to execute when state is exited
    """

    def __init__(self, name: str, on_entry: str = None, on_exit: str = None) -> None:
        ContractMixin.__init__(self)
        StateMixin.__init__(self, name)
        ActionStateMixin.__init__(self, on_entry, on_exit)

    def __eq__(self, other):
        if isinstance(other, FinalState):
            return (
                ContractMixin.__eq__(self, other)
                and StateMixin.__eq__(self, other)
                and ActionStateMixin.__eq__(self, other)
            )
        else:
            return NotImplemented


class Transition(ContractMixin):
    """
    Represent a transition from a source state to a target state.

    A transition can be eventless (no event) or internal (no target).
    A condition (code as string) can be specified as a guard.

    :param source: name of the source state
    :param target: name of the target state (if transition is not internal)
    :param event: event name (if any)
    :param guard: condition as code (if any)
    :param action: action as code (if any)
    :param priority: priority (default to 0)
    """

    LOW_PRIORITY = -1
    DEFAULT_PRIORITY = 0
    HIGH_PRIORITY = 1

    def __init__(self, source: str, target: str = None, event: str = None, guard: str = None,
                 action: str = None, priority=None) -> None:
        ContractMixin.__init__(self)
        self._source = source
        self._target = target
        self.event = event
        self.guard = guard
        self.action = action
        self.priority = 0 if priority is None else priority

    @property
    def source(self):
        return self._source

    @property
    def target(self):
        return self._target

    @property
    def internal(self):
        """
        Boolean indicating whether this transition is an internal transition.
        """
        return self._target is None

    @property
    def eventless(self):
        """
        Boolean indicating whether this transition is an eventless transition.
        """
        return self.event is None

    def __eq__(self, other):
        if isinstance(other, Transition):
            return (
                ContractMixin.__eq__(self, other)
                and self.source == other.source
                and self.target == other.target
                and self.event == other.event
                and self.guard == other.guard
                and self.action == other.action
                and self.priority == other.priority
            )
        else:
            return NotImplemented

    def __repr__(self):
        return 'Transition({!r}, {!r}, event={!r})'.format(self.source, self.target, self.event)

    def __str__(self):
        return '{} -> {}{} [{}] -> {}'.format(
            self.source,
            '' if self.priority == 0 else '{}:'.format(self.priority),
            self.event,
            self.guard,
            self.target if self.target else ''
        )

    def __hash__(self):
        return hash(self.source)



import ruamel.yaml as yaml
import schema

from ..exceptions import StatechartError
from ..model import Statechart

from .datadict import export_to_dict, import_from_dict

__all__ = ['import_from_yaml', 'export_to_yaml']


class SCHEMA:
    contract = {schema.Or('before', 'after', 'always'): schema.Use(str)}

    transition = {
        schema.Optional('target'): schema.Use(str),
        schema.Optional('event'): schema.Use(str),
        schema.Optional('guard'): schema.Use(str),
        schema.Optional('action'): schema.Use(str),
        schema.Optional('contract'): [contract],
        schema.Optional('priority'): schema.Or(schema.Use(int), 'high', 'low'),
    }

    state = dict()  # type: ignore
    state.update({
        'name': schema.Use(str),
        schema.Optional('type'): schema.Or('final', 'shallow history', 'deep history'),
        schema.Optional('on entry'): schema.Use(str),
        schema.Optional('on exit'): schema.Use(str),
        schema.Optional('transitions'): [transition],
        schema.Optional('contract'): [contract],
        schema.Optional('initial'): schema.Use(str),
        schema.Optional('parallel states'): [state],
        schema.Optional('states'): [state],
        schema.Optional('memory'): schema.Use(str),
    })

    statechart = {
        'statechart': {
            'name': schema.Use(str),
            schema.Optional('description'): schema.Use(str),
            schema.Optional('preamble'): schema.Use(str),
            'root state': state,
        }
    }


def import_from_yaml(
        text: str = None, filepath: str = None, *, ignore_schema: bool = False,
        ignore_validation: bool = False) -> Statechart:
    """
    Import a statechart from a YAML representation (first argument) or a YAML file (filepath
    argument).

    Unless specified, the structure contained in the YAML is validated against a predefined
    schema (see *sismic.io.SCHEMA*), and the resulting statechart is validated using its
    *validate()* method.

    :param text: A YAML text. If not provided, filepath argument has to be provided.
    :param filepath: A path to a YAML file.
    :param ignore_schema: set to *True* to disable yaml validation.
    :param ignore_validation: set to *True* to disable statechart validation.
    :return: a *Statechart* instance
    """
    if not text and not filepath:
        raise TypeError(
            'A YAML must be provided, either using first argument or filepath argument.')
    elif text and filepath:
        raise TypeError('Either provide first argument or filepath argument, not both.')
    elif filepath:
        with open(filepath, 'r') as f:
            text = f.read()

    if yaml.version_info < (0, 15):
        data = yaml.safe_load(text)  # type: dict
    else:
        yml = yaml.YAML(typ='safe', pure=True)
        data = yml.load(text)

    if not ignore_schema:
        try:
            data = schema.Schema(SCHEMA.statechart).validate(data)
        except schema.SchemaError as e:
            raise StatechartError('YAML validation failed') from e

    sc = import_from_dict(data)

    if not ignore_validation:
        sc.validate()
    return sc


def export_to_yaml(statechart: Statechart, filepath: str = None) -> str:
    """
    Export given *Statechart* instance to YAML. Its YAML representation is returned by
    this function. Automatically save the output to filepath, if provided.

    :param statechart: statechart to export
    :param filepath: save output to given filepath, if provided
    :return: A textual YAML representation
    """
    output = yaml.dump(export_to_dict(statechart, ordered=False),
                       width=1000, default_flow_style=False)

    if filepath:
        with open(filepath, 'w') as f:
            f.write(output)

    return output



from typing import Callable, Any

from ..model import MetaEvent, Event

from ..exceptions import PropertyStatechartError


__all__ = ['InternalEventListener', 'PropertyStatechartListener']


class InternalEventListener:
    """
    Listener that filters and propagates internal events as external events.
    """

    def __init__(self, callable: Callable[[Event], Any]) -> None:
        self._callable = callable

    def __call__(self, event: MetaEvent) -> None:
        if event.name == 'event sent':
            self._callable(Event(event.event.name, **event.event.data))


class PropertyStatechartListener:
    """
    Listener that propagates meta-events to given property statechart, executes
    the property statechart, and checks it.
    """

    def __init__(self, interpreter) -> None:
        self._interpreter = interpreter

    def __call__(self, event: MetaEvent) -> None:
        self._interpreter.queue(event)
        self._interpreter.execute()
        if self._interpreter.final:
            raise PropertyStatechartError(self._interpreter)



import pytest

from sismic.interpreter import Event, MetaEvent, InternalEvent
from sismic.exceptions import PropertyStatechartError


class TestInterpreterMetaEvents:
    @pytest.fixture
    def property_statechart(self, microwave, mocker):
        prop_sc = mocker.MagicMock(name='Interpreter', spec=microwave)
        prop_sc.queue = mocker.MagicMock(return_value=None)
        prop_sc.execute = mocker.MagicMock(return_value=None)
        prop_sc.final = False
        prop_sc.time = 0
        
        def prop_sc_call(statechart, clock):    
            return prop_sc

        # Bind it
        microwave.bind_property_statechart(None, interpreter_klass=prop_sc_call)

        return prop_sc

    def test_deprecated_interpreter(self, microwave):
        with pytest.warns(DeprecationWarning):
            microwave.bind_property_statechart(microwave)

    def test_synchronised_time(self, microwave, property_statechart):
        assert microwave.time == property_statechart.time
        microwave.clock.time += 10
        assert microwave.time == property_statechart.time

    def test_empty_step(self, microwave, property_statechart):
        microwave.execute()
        assert property_statechart.queue.call_args_list[0][0][0] == MetaEvent('step started', time=0)
        assert property_statechart.queue.call_args_list[-1][0][0] == MetaEvent('step ended')

        for call in property_statechart.queue.call_args_list:
            assert isinstance(call[0][0], MetaEvent)

    def test_event_sent(self, microwave, property_statechart):
        # Add send to a state
        state = microwave.statechart.state_for('door closed')
        state.on_entry = 'send("test")'

        microwave.execute()

        assert MetaEvent('event sent', event=InternalEvent('test')) in [x[0][0] for x in property_statechart.queue.call_args_list]

    def test_meta_event_sent(self, microwave, property_statechart):
        # Add meta to a state
        state = microwave.statechart.state_for('door closed')
        state.on_entry = 'notify("test", x=1, y="hello")'

        microwave.execute()

        assert MetaEvent('event sent', event=MetaEvent('test', x=1, y='hello')) not in [x[0][0] for x in property_statechart.queue.call_args_list]
        assert MetaEvent('test', x=1, y='hello') in [x[0][0] for x in property_statechart.queue.call_args_list]

    def test_trace(self, microwave, property_statechart):
        microwave.queue('door_opened')
        microwave.execute()

        call_list = [
            MetaEvent('step started', time=0),
            MetaEvent('state entered', state='controller'),
            MetaEvent('state entered', state='door closed'),
            MetaEvent('state entered', state='closed without item'),
            MetaEvent('step ended'),

            MetaEvent('step started', time=0),
            MetaEvent('event consumed', event=Event('door_opened')),
            MetaEvent('state exited', state='closed without item'),
            MetaEvent('state exited', state='door closed'),
            MetaEvent('transition processed', source='closed without item', target='opened without item', event=Event('door_opened')),
            MetaEvent('state entered', state='door opened'),
            MetaEvent('state entered', state='opened without item'),
            MetaEvent('event sent', event=InternalEvent('lamp_switch_on')),
            MetaEvent('step ended')
        ]

        for i, call in enumerate(call_list):
            effective_call = property_statechart.queue.call_args_list[i][0][0]

            assert isinstance(effective_call, MetaEvent)
            assert call.name == effective_call.name

            for param in ['state', 'source', 'target', 'event']:
                if param in call.data:
                    assert effective_call.data[param] == call.data[param]

    def test_final(self, microwave, property_statechart):
        microwave.execute()
        assert not property_statechart.final

        property_statechart.final = True
        assert property_statechart.final

        with pytest.raises(PropertyStatechartError):
            microwave.execute()



import unittest
from sismic.io import import_from_yaml
from sismic.interpreter import Interpreter


class MicrowaveTests(unittest.TestCase):
    def setUp(self):
        with open('microwave.yaml') as f:
            sc = import_from_yaml(f)

        self.oven = Interpreter(sc)
        self.oven.execute_once()

    def test_no_heating_when_door_is_not_closed(self):
        self.oven.queue('door_opened', 'item_placed', 'timer_inc')
        self.oven.execute()

        self.oven.queue('cooking_start')
        
        for step in iter(self.oven.execute_once, None):
            for event in step.sent_events:
                self.assertNotEqual(event.name, 'heating_on')
        
        self.assertNotIn('cooking_mode', self.oven.configuration)

    def test_increase_timer(self):
        self.oven.queue('door_opened', 'item_placed', 'door_closed')
    
        events = 10 * ['timer_inc']    
        self.oven.queue(*events)
        self.oven.execute()
        
        self.assertEqual(self.oven.context['timer'], 10)



import abc
from typing import Any, Optional, Iterable, List, Mapping

from ..model import Statechart, StateMixin, Transition, Event
from ..exceptions import CodeEvaluationError

__all__ = ['Evaluator']


class Evaluator(metaclass=abc.ABCMeta):
    """
    Abstract base class for any evaluator.

    An instance of this class defines what can be done with piece of codes
    contained in a statechart (condition, action, etc.).

    Notice that the execute_* methods are called at each step, even if there is no
    code to execute. This allows the evaluator to keep track of the states that are
    entered or exited, and of the transitions that are processed.

    :param interpreter: the interpreter that will use this evaluator,
        is expected to be an *Interpreter* instance
    :param initial_context: an optional dictionary to populate the context
    """

    @abc.abstractmethod
    def __init__(self, interpreter=None, *, initial_context: Mapping[str, Any] = None) -> None:
        pass

    @property
    @abc.abstractmethod
    def context(self) -> Mapping[str, Any]:
        """
        The context of this evaluator. A context is a dict-like mapping between
        variables and values that is expected to be exposed when the code is evaluated.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _evaluate_code(self, code: str, *, additional_context: Mapping[str, Any] = None) -> bool:
        """
        Generic method to evaluate a piece of code. This method is a fallback if one of
        the other evaluate_* methods is not overridden.

        :param code: code to evaluate
        :param additional_context: an optional additional context
        :return: truth value of *code*
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _execute_code(
            self, code: str, *, additional_context: Mapping[str, Any] = None) -> List[Event]:
        """
        Generic method to execute a piece of code. This method is a fallback if one
        of the other execute_* methods is not overridden.

        :param code: code to execute
        :param additional_context: an optional additional context
        :return: a list of sent events
        """
        raise NotImplementedError()

    def execute_statechart(self, statechart: Statechart):
        """
        Execute the initial code of a statechart.
        This method is called at the very beginning of the execution.

        :param statechart: statechart to consider
        """
        if statechart.preamble:
            events = self._execute_code(statechart.preamble)
            if len(events) > 0:
                raise CodeEvaluationError('Events cannot be raised by statechart preamble')

    def evaluate_guard(
            self, transition: Transition, event: Optional[Event] = None) -> Optional[bool]:
        """
        Evaluate the guard for given transition.

        :param transition: the considered transition
        :param event: instance of *Event* if any
        :return: truth value of *code*
        """
        if transition.guard:
            return self._evaluate_code(transition.guard, additional_context={'event': event})
        return None

    def execute_action(self, transition: Transition, event: Optional[Event] = None) -> List[Event]:
        """
        Execute the action for given transition.
        This method is called for every transition that is processed, even those with no *action*.

        :param transition: the considered transition
        :param event: instance of *Event* if any
        :return: a list of sent events
        """
        if transition.action:
            return self._execute_code(transition.action, additional_context={'event': event})
        else:
            return []

    def execute_on_entry(self, state: StateMixin) -> List[Event]:
        """
        Execute the on entry action for given state.
        This method is called for every state that is entered, even those with no *on_entry*.

        :param state: the considered state
        :return: a list of sent events
        """
        code = getattr(state, 'on_entry', None)
        if code:
            return self._execute_code(code)
        else:
            return []

    def execute_on_exit(self, state: StateMixin) -> List[Event]:
        """
        Execute the on exit action for given state.
        This method is called for every state that is exited, even those with no *on_exit*.

        :param state: the considered state
        :return: a list of sent events
        """
        code = getattr(state, 'on_exit', None)
        if code:
            return self._execute_code(code)
        else:
            return []

    def evaluate_preconditions(self, obj, event: Optional[Event] = None) -> Iterable[str]:
        """
        Evaluate the preconditions for given object (either a *StateMixin* or a
        *Transition*) and return a list of conditions that are not satisfied.

        :param obj: the considered state or transition
        :param event: an optional *Event* instance, if any
        :return: list of unsatisfied conditions
        """
        event_d = {'event': event} if isinstance(obj, Transition) else None
        return filter(
            lambda c: not self._evaluate_code(
                c, additional_context=event_d), getattr(obj, 'preconditions', [])
        )

    def evaluate_invariants(self, obj, event: Optional[Event] = None) -> Iterable[str]:
        """
        Evaluate the invariants for given object (either a *StateMixin* or a
        *Transition*) and return a list of conditions that are not satisfied.

        :param obj: the considered state or transition
        :param event: an optional *Event* instance, if any
        :return: list of unsatisfied conditions
        """
        event_d = {'event': event} if isinstance(obj, Transition) else None
        return filter(
            lambda c: not self._evaluate_code(
                c, additional_context=event_d), getattr(obj, 'invariants', [])
        )

    def evaluate_postconditions(self, obj, event: Optional[Event] = None) -> Iterable[str]:
        """
        Evaluate the postconditions for given object (either a *StateMixin* or a
        *Transition*) and return a list of conditions that are not satisfied.

        :param obj: the considered state or transition
        :param event: an optional *Event* instance, if any
        :return: list of unsatisfied conditions
        """
        event_d = {'event': event} if isinstance(obj, Transition) else None
        return filter(
            lambda c: not self._evaluate_code(
                c, additional_context=event_d), getattr(obj, 'postconditions', [])
        )
